/* This file is part of ESS++.
 *      Copyright (c) Marc Chadeau (m.chadeau@imperial.ac.uk)
 *                    Leonardo Bottolo (l.bottolo@imperial.ac.uk)
 *                    David Hastie (d.hastie@imperial.ac.uk)
 *      2010
 *
 * ESS++ is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ESS++ is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ESS++.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "cond_post.h"
#include "struc.h"
#include <stdlib.h>
#define DEBUG 0

using namespace std;

void get_vect_gam_init(vector < vector <unsigned int> > &vect_gam,
		       vector < vector <unsigned int> > &Gam_step_regr,
		       bool iso_T_Flag)
{

  if(!iso_T_Flag){
    for(unsigned int chain=0;chain<vect_gam.size();chain++){
      for(unsigned int row=0;row<Gam_step_regr.size();row++){
	for(unsigned int col=0;col<Gam_step_regr[row].size();col++){
	  unsigned int temp_pos=Gam_step_regr[row][col];
	  vect_gam[chain][temp_pos]=1;
	} 
      }
    }
  }
  else{
    unsigned int *source_list= new unsigned int [vect_gam[0].size()];
    for(unsigned int row=0;row<Gam_step_regr.size();row++){
      for(unsigned int col=0;col<Gam_step_regr[row].size();col++){
	unsigned int temp_pos=Gam_step_regr[row][col];
	vect_gam[0][temp_pos]=1;
      }
    }
    //Counting the number of Vars in
    unsigned int n_vars_in=0;
    for(unsigned int col=0;col<vect_gam[0].size();col++){
      if(vect_gam[0][col]==1){
	n_vars_in++;
      }
      source_list[col]=col;
    }
    cout << endl;
    //Filling the other chains randomly
    //Case 1: No vars in: sample one var for chains #2 to C
    if(n_vars_in==0){
      for(unsigned int chain=1;chain<vect_gam.size();chain++){
	unsigned int *list_sample = new unsigned int [1];
	My_gsl_ran_choose_u_int(list_sample,
				1,
				source_list,
				vect_gam[0].size());
	unsigned int curr_pos=list_sample[0];
	vect_gam[chain][curr_pos]=1;
	delete [] list_sample;
      }
    } 
    else{//if n_vars_in>0: sample n_vars in for each chain #2 to C
      for(unsigned int chain=1;chain<vect_gam.size();chain++){
	unsigned int *list_sample = new unsigned int [n_vars_in];
	My_gsl_ran_choose_u_int(list_sample,
				n_vars_in,
				source_list,
				vect_gam[0].size());
	for(unsigned int curr_var=0;curr_var<n_vars_in;curr_var++){
	  unsigned int curr_pos=list_sample[curr_var];
	  vect_gam[chain][curr_pos]=1;
	}
	delete [] list_sample;
      }
    }
    delete [] source_list;
  }
}

void get_mat_Q_and_R(gsl_matrix *mat_X_gam,
		     gsl_matrix *mat_Q,
		     gsl_matrix *mat_R,
		     gsl_vector *tau)
{
  

  //Declaring the vectors for QR decomposition


 //Step 1. QR decomposition
  gsl_linalg_QR_decomp(mat_X_gam,tau);
  
  //Step 2. Extracting the full Q and R matrices
  gsl_linalg_QR_unpack(mat_X_gam,
		       tau,
		       mat_Q,
		       mat_R);
}

gsl_matrix *get_RSP_g_prior(gsl_matrix *mat_Q,
			    gsl_matrix *mat_Y,
			    unsigned int p_X_gam)
{
  //This function is only called if p_X_gam (# vars in) <= nX (# obs)
  //Otherwise RSP is a null matrix pYxpY (ie # outcome)
  
  //Defining Q_sub
  unsigned int n = mat_Q->size1;
  gsl_matrix *mat_RSP=gsl_matrix_calloc(mat_Y->size2,mat_Y->size2); 

  if(p_X_gam<mat_Q->size2){
    unsigned int n_col_Q_sub=mat_Q->size2-(p_X_gam);
    
    gsl_matrix_view viewMatQ = gsl_matrix_submatrix(mat_Q,0,p_X_gam,n,n_col_Q_sub);
    
    gsl_matrix* Q_sub=&(viewMatQ.matrix);

    gsl_matrix *Q_subT_Y=gsl_matrix_alloc(Q_sub->size2,mat_Y->size2); 

    gsl_blas_dgemm(CblasTrans,
		   CblasNoTrans,
		   1.0,
		   Q_sub,
		   mat_Y,
		   0.0,
		   Q_subT_Y);

    gsl_blas_dgemm(CblasTrans,
		   CblasNoTrans,
		   1.0,
		   Q_subT_Y,
		   Q_subT_Y,
		   0.0,
		   mat_RSP);

    gsl_matrix_free(Q_subT_Y);

  }
  
  return mat_RSP;

}

double get_prior_gam(Prior_param PR,
		     unsigned int nb_var_in,
		     unsigned int pX)
{
  double temp1=gsl_sf_lngamma(PR.a_pi+(double)(nb_var_in));
  double temp2=gsl_sf_lngamma((double)(pX)+PR.b_pi-(double)(nb_var_in));
  double temp3=gsl_sf_lngamma((double)(pX)+PR.b_pi+PR.a_pi);
  double res=temp1 + temp2 - temp3;

  return res;

}		   

double get_cond_posterior(Prior_param PR,
			  int g_sample,
			  int g_opt,
			  double g)
{
  double res=0.0;
  if(g_sample!=0){
    if(g_opt==1){//g-prior
       res=LB_inv_gamma_pdf(g,
			   PR.alpha,
			   PR.beta);

    }
    else{//ind prior
      LB_inv_exp_pdf(g,
		     PR.mu);
    }

  }
  else{
    res=0.0;
  }
  return res;
}

double LB_inv_gamma_pdf(double x,
			double alpha,
			double beta)
{

  double temp1 = alpha * log(beta) - gsl_sf_lngamma(alpha);
  double temp2 = -(alpha + 1) * log(x) -beta/x;
  double res=log(exp(temp1 + temp2));

  return res;

}

double LB_inv_exp_pdf(double x,
		      double mu)
{
  
  double z = x/mu;
  double res=exp(-z)/mu;
  return log(res);
  
}

double get_log_Marg(Prior_param PR,
		    gsl_matrix *mat_Y,
		    gsl_matrix *mat_RSP,
		    double g,
		    unsigned int n_var_in,
		    unsigned int nX,
		    unsigned int pY,
		    int g_opt)
{
  //Building Qk
  double res=0.0;

  gsl_matrix *Qk=gsl_matrix_calloc(pY,pY);
  for(unsigned int row=0;row<pY;row++){
    Qk->data[row*(1+Qk->size2)]=PR.k;
  }

  //Calculating Qstar
  gsl_matrix *Qstar=gsl_matrix_calloc(pY,pY);

  if(g_opt==1){
    //Step 1: getting A=Y'.Y/(g+1)
    gsl_blas_dgemm(CblasTrans,
		   CblasNoTrans,
		   1.0/(g+1.0),
		   mat_Y,
		   mat_Y,
		   0.0,
		   Qstar);

    //Step 2: get B=(g/(g+1))*RSP
    double K=g/(g+1);
    gsl_matrix_scale(mat_RSP,K);

    //Step 3: summing A and B: Qgam
    gsl_matrix_add(Qstar,mat_RSP);

    //Step4: Qstar=Qgam + QK
    gsl_matrix_add(Qstar,Qk);

    gsl_permutation * p = gsl_permutation_alloc (pY);
    int s=0;
    
    gsl_linalg_LU_decomp (Qstar, p, &s);
    double det_Qstar=gsl_linalg_LU_det (Qstar,s);

    gsl_permutation_free (p);
    res=(-((double)(pY)/2.0)*
	 (double)(n_var_in) *
	 log(g+1)-
	 ((PR.delta+(double)(nX)+(double)(pY)-1.0) / 2.0) *
	 log(det_Qstar)); 
  }
  else{
    //Independent prior not yet coded
  }

  gsl_matrix_free(Qk);
  gsl_matrix_free(Qstar);

  return res;


}

void full_log_marg_computation(Double_Matrices mat_log_marg,
			       Double_Matrices mat_log_cond_post,
			       gsl_matrix *mat_X_gam,
			       gsl_matrix *mat_Y,
			       Prior_param PR,
			       int g_opt,
			       int g_sample,
			       double g,
			       unsigned int pX,
			       unsigned int current_chain,
			       unsigned int current_sweep,
		               bool cudaFlag)
{
 
  unsigned int pX_gam=mat_X_gam->size2;
  unsigned int pY=mat_Y->size2;
  unsigned int nX=mat_X_gam->size1;

  double logMargLik;
  gsl_vector *tau=gsl_vector_alloc(GSL_MIN(mat_X_gam->size1,mat_X_gam->size2));
  gsl_matrix *mat_Q=gsl_matrix_alloc(mat_X_gam->size1,mat_X_gam->size1);
  gsl_matrix *mat_R=gsl_matrix_alloc(mat_X_gam->size1,mat_X_gam->size2);

  get_mat_Q_and_R(mat_X_gam,
		  mat_Q,
		  mat_R,
		  tau);
  

  gsl_matrix *mat_RSP=get_RSP_g_prior(mat_Q,
				      mat_Y,
				      mat_X_gam->size2);

  logMargLik=get_log_Marg(PR,
                            mat_Y,
                            mat_RSP,
                            g,
                            pX_gam,
                            nX,
                            pY,
                            g_opt);
  gsl_matrix_free(mat_Q);
  gsl_matrix_free(mat_R);
  gsl_vector_free(tau);
  gsl_matrix_free(mat_RSP);


  double log_p_gam=get_prior_gam(PR,
				 pX_gam,
				 pX);

  double log_p_g=get_cond_posterior(PR,
				    g_sample,
				    g_opt,
				    g);


  double log_cond_post=log_p_gam + log_p_g + logMargLik;
 
  mat_log_marg.matrix[current_chain][current_sweep]=logMargLik;
  mat_log_cond_post.matrix[current_chain][current_sweep]=log_cond_post;

}

void full_log_marg_computation_no_var_in(Double_Matrices mat_log_marg,
					 Double_Matrices mat_log_cond_post,
					 gsl_matrix *mat_Y,
					 Prior_param PR,
					 int g_opt,
					 int g_sample,
					 double g,
					 unsigned int pX,
					 unsigned int current_chain,
					 unsigned int current_sweep)
{
 
  unsigned int pY=mat_Y->size2;
  unsigned int nX=mat_Y->size1;

  gsl_matrix *mat_Q=gsl_matrix_alloc(nX,nX);
  

  gsl_matrix_set_identity(mat_Q);


  gsl_matrix *mat_RSP=get_RSP_g_prior(mat_Q,
				      mat_Y,
				      0);

  double log_p_gam=get_prior_gam(PR,
				 0,
				 pX);

  double log_p_g=get_cond_posterior(PR,
				    g_sample,
				    g_opt,
				    g);

  double log_marg=get_log_Marg(PR,
			       mat_Y,
			       mat_RSP,
			       g,
			       0,
			       nX,
			       pY,
			       g_opt);

  double log_cond_post=log_p_gam + log_p_g + log_marg;
 
  mat_log_marg.matrix[current_chain][current_sweep]=log_marg;
  mat_log_cond_post.matrix[current_chain][current_sweep]=log_cond_post;

  gsl_matrix_free(mat_Q);
  gsl_matrix_free(mat_RSP);
}

void get_log_marg_cond_post_only(gsl_vector *Result,
				 gsl_matrix *mat_X_gam,
				 gsl_matrix *mat_Y,
				 Prior_param PR,
				 int g_opt,
				 int g_sample,
				 double g,
				 unsigned int pX,
		                 bool cudaFlag)
{
 
  unsigned int pX_gam=mat_X_gam->size2;
  unsigned int pY=mat_Y->size2;
  unsigned int nX=mat_X_gam->size1;
  
  double logMargLik;
  gsl_vector *tau=gsl_vector_alloc(GSL_MIN(mat_X_gam->size1,mat_X_gam->size2));
  gsl_matrix *mat_Q=gsl_matrix_alloc(mat_X_gam->size1,mat_X_gam->size1);
  gsl_matrix *mat_R=gsl_matrix_alloc(mat_X_gam->size1,mat_X_gam->size2);

  get_mat_Q_and_R(mat_X_gam,
                    mat_Q,
                    mat_R,
                    tau);
  
  
  
  gsl_matrix *mat_RSP=get_RSP_g_prior(mat_Q,
                                        mat_Y,
                                        mat_X_gam->size2);

  logMargLik = get_log_Marg(PR,
                              mat_Y,
                              mat_RSP,
                              g,
                              pX_gam,
                              nX,
                              pY,
                              g_opt);

  gsl_matrix_free(mat_Q);
  gsl_matrix_free(mat_R);
  gsl_vector_free(tau);
  gsl_matrix_free(mat_RSP);


  double log_p_gam=get_prior_gam(PR,
				 pX_gam,
				 pX);

  double log_p_g=get_cond_posterior(PR,
				    g_sample,
				    g_opt,
				    g);

  Result->data[0]=logMargLik;
  Result->data[1]=log_p_gam + log_p_g + logMargLik;



}

void get_log_marg_cond_post_only_no_var_in(gsl_vector *Result,
					   gsl_matrix *mat_Y,
					   Prior_param PR,
					   int g_opt,
					   int g_sample,
					   double g,
					   unsigned int pX)
{
 
  unsigned int pY=mat_Y->size2;
  unsigned int nX=mat_Y->size1;

  gsl_matrix *mat_Q=gsl_matrix_alloc(nX,nX);
  
  gsl_matrix_set_identity(mat_Q);
 
  gsl_matrix *mat_RSP=get_RSP_g_prior(mat_Q,
				      mat_Y,
				      0);

  double log_p_gam=get_prior_gam(PR,
				 0,
				 pX);

  double log_p_g=get_cond_posterior(PR,
				    g_sample,
				    g_opt,
				    g);

  Result->data[0]=get_log_Marg(PR,
			       mat_Y,
			       mat_RSP,
			       g,
			       0,
			       nX,
			       pY,
			       g_opt);
  
  Result->data[1]=log_p_gam + log_p_g + Result->data[0];
 
  gsl_matrix_free(mat_Q);
  gsl_matrix_free(mat_RSP);
  
  
}
