/* This file is part of ESS++.
 *      Copyright (c) Marc Chadeau (m.chadeau@imperial.ac.uk)
 *                    Leonardo Bottolo (l.bottolo@imperial.ac.uk)
 *                    David Hastie (d.hastie@imperial.ac.uk)
 *      2010
 *
 * ESS++ is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ESS++ is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ESS++.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef COND_POST_H
#define COND_POST_H
#include <gsl/gsl_fit.h>
#include <gsl/gsl_cdf.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_sort.h>
#include <gsl/gsl_sort_vector.h>
#include <gsl/gsl_permutation.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_vector.h>
#include "../Classes/Double_Matrices.h"
#include "../Classes/Double_Matrices_cont.h"
#include "../Classes/Int_Matrices.h"
#include "../Routines/matrix_handling.h"
#include "../Routines/rand.h"
#include <vector>
#include <gsl/gsl_math.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_linalg.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <cstring>
#include <iostream>
#include <sstream> 
#include <cstdarg>
#include <math.h>
#include <cmath>
#include <math.h>
#include <ctype.h>


#include <gsl/gsl_statistics_double.h>
#include <gsl/gsl_sf_gamma.h>
#include "../Classes/Prior_param.h"

void get_vect_gam_init(vector < vector <unsigned int> > &vect_gam,
		       vector < vector <unsigned int> > &Gam_step_regr,
		       bool iso_T_Flag);

void get_mat_Q_and_R(gsl_matrix *mat_X_gam,
		     gsl_matrix *mat_Q,
		     gsl_matrix *mat_R,
		     gsl_vector *tau);

gsl_matrix *get_RSP_g_prior(gsl_matrix *mat_Q,
			    gsl_matrix *mat_Y,
			    unsigned int p_X_gam);

double get_prior_gam(Prior_param PR,
		     unsigned int nb_var_in,
		     unsigned int pX);

double get_cond_posterior(Prior_param PR,
			  int g_sample,
			  int g_opt,
			  double g);

double LB_inv_gamma_pdf(double x,
			double alpha,
			double beta);

double LB_inv_exp_pdf(double x,
		      double mu);

double get_log_Marg(Prior_param PR,
		    gsl_matrix *mat_Y,
		    gsl_matrix *mat_RSP,
		    double g,
		    unsigned int n_var_in,
		    unsigned int nX,
		    unsigned int pY,
		    int g_opt);

void full_log_marg_computation(Double_Matrices mat_log_marg,
			       Double_Matrices mat_log_cond_post,
			       gsl_matrix *mat_X_gam,
			       gsl_matrix *mat_Y,
			       Prior_param PR,
			       int g_opt,
			       int g_sample,
			       double g,
			       unsigned int pX,
			       unsigned int current_chain,
			       unsigned int current_sweep,
		               bool cudaFlag);

void full_log_marg_computation_no_var_in(Double_Matrices mat_log_marg,
					 Double_Matrices mat_log_cond_post,
					 gsl_matrix *mat_Y,
					 Prior_param PR,
					 int g_opt,
					 int g_sample,
					 double g,
					 unsigned int pX,
					 unsigned int current_chain,
					 unsigned int current_sweep);

void get_log_marg_cond_post_only(gsl_vector *Result,
				 gsl_matrix *mat_X_gam,
				 gsl_matrix *mat_Y,
				 Prior_param PR,
				 int g_opt,
				 int g_sample,
				 double g,
				 unsigned int pX,
		                 bool cudaFlag);

void get_log_marg_cond_post_only_no_var_in(gsl_vector *Result,
					   gsl_matrix *mat_Y,
					   Prior_param PR,
					   int g_opt,
					   int g_sample,
					   double g,
					   unsigned int pX);
#endif /*  */
